use std::io::{Read, Write};
use std::{fs, path};

use crate::domain::tls::PrivateKey;
use crate::error::unexpected::{self, Mappable};
use crate::util;

#[mockall::automock]
pub trait Store {
    fn set(&self, k: &PrivateKey) -> Result<(), unexpected::Error>;
    fn get(&self) -> unexpected::Result<Option<PrivateKey>>;
}

pub struct FSStore {
    dir_path: path::PathBuf,
}

impl FSStore {
    pub fn new(dir_path: &path::Path) -> Result<Self, unexpected::Error> {
        fs::create_dir_all(dir_path).or_unexpected()?;
        Ok(Self {
            dir_path: dir_path.into(),
        })
    }

    fn account_key_path(&self) -> path::PathBuf {
        self.dir_path.join("account.key")
    }
}

impl Store for FSStore {
    fn set(&self, k: &PrivateKey) -> Result<(), unexpected::Error> {
        let path = self.account_key_path();
        {
            let mut file = fs::File::create(&path).or_unexpected_while("creating file")?;
            file.write_all(k.to_string().as_bytes())
                .or_unexpected_while("writing file")
        }
        .map_unexpected_while(|| format!("path is {}", path.display()))?;
        Ok(())
    }

    fn get(&self) -> unexpected::Result<Option<PrivateKey>> {
        let path = self.account_key_path();
        {
            let mut file =
                match util::open_file(path.as_path()).or_unexpected_while("opening_file")? {
                    Some(file) => file,
                    None => return Ok(None),
                };

            let mut key = String::new();
            file.read_to_string(&mut key)
                .or_unexpected_while("reading file")?;

            let key: PrivateKey = key.parse().or_unexpected_while("parsing private key")?;

            unexpected::Result::<Option<PrivateKey>>::Ok(Some(key))
        }
        .map_unexpected_while(|| format!("path is {}", path.display()))
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::domain::tls;
    use tempdir::TempDir;

    #[test]
    fn account_key() {
        let tmp_dir = TempDir::new("domain_acme_account_key_store").unwrap();
        let store = FSStore::new(tmp_dir.path()).expect("store created");

        assert!(matches!(
            store.get(),
            unexpected::Result::<Option<PrivateKey>>::Ok(None)
        ));

        let k = tls::PrivateKey::new();

        store.set(&k).expect("account private key set");

        assert_eq!(Some(k), store.get().expect("account private key retrieved"));
    }
}
