use std::net;
use std::str::FromStr;

use crate::error::unexpected::{self, Mappable};
use crate::{domain, token};

use trust_dns_client::client::{AsyncClient, ClientHandle};
use trust_dns_client::rr::{DNSClass, Name, RData, RecordType};
use trust_dns_client::udp;

use rand::Rng;

#[derive(thiserror::Error, Debug)]
pub enum CheckDomainError {
    #[error("no service dns records set")]
    ServiceDNSRecordsNotSet,

    #[error("challenge token not set")]
    ChallengeTokenNotSet,

    #[error(transparent)]
    Unexpected(#[from] unexpected::Error),
}

pub enum DNSRecord {
    A(net::Ipv4Addr),
    AAAA(net::Ipv6Addr),
    CNAME(domain::Name),
}

pub struct DNSChecker {
    // TODO we should use some kind of connection pool here, I suppose
    client: tokio::sync::Mutex<AsyncClient>,
    token_store: Box<dyn token::Store + Send + Sync>,
}

impl DNSChecker {
    pub async fn new<TokenStore>(
        token_store: TokenStore,
        config: &domain::ConfigDNS,
    ) -> Result<Self, unexpected::Error>
    where
        TokenStore: token::Store + Send + Sync + 'static,
    {
        let stream = udp::UdpClientStream::<tokio::net::UdpSocket>::new(config.resolver_addr);
        let (client, bg) = AsyncClient::connect(stream).await.or_unexpected()?;
        tokio::spawn(bg);
        // TODO there should be a mechanism to clean this up

        Ok(Self {
            token_store: Box::from(token_store),
            client: tokio::sync::Mutex::new(client),
        })
    }

    pub fn get_challenge_token(&self, domain: &domain::Name) -> unexpected::Result<Option<String>> {
        self.token_store.get(domain.as_str())
    }

    pub async fn check_domain(
        &self,
        domain: &domain::Name,
        challenge_token: &str,
    ) -> Result<(), CheckDomainError> {
        let domain_rr = domain.as_rr();

        // check that the TXT record with the challenge token is correctly installed on the domain
        {
            let domain = Name::from_str("_domani_challenge")
                .or_unexpected_while("parsing TXT name")?
                .append_domain(domain_rr)
                .or_unexpected_while("appending domain to TXT")?;

            let response = self
                .client
                .lock()
                .await
                .query(domain, DNSClass::IN, RecordType::TXT)
                .await
                .or_unexpected_while("querying TXT record")?;

            let records = response.answers();

            if !records.iter().any(|record| -> bool {
                match record.data() {
                    Some(RData::TXT(txt)) => txt.to_string().contains(challenge_token),
                    _ => false,
                }
            }) {
                return Err(CheckDomainError::ChallengeTokenNotSet);
            }
        }

        // check that DNS correctly resolves for the domain. This is done by serving an HTTP
        // challenge on the domain, which we then query for here.
        //
        // first store the challenge token, so that the HTTP server can find it via
        // get_challenge_token.
        let token: String = rand::thread_rng()
            .sample_iter(rand::distributions::Alphanumeric)
            .take(16)
            .map(char::from)
            .collect();

        self.token_store
            .set(domain.as_str().to_string(), token.clone())
            .or_unexpected_while("storing challenge token")?;

        let body = match reqwest::get(format!(
            "http://{}/.well-known/domani-challenge",
            domain.as_str(),
        ))
        .await
        {
            Err(_) => return Err(CheckDomainError::ServiceDNSRecordsNotSet),
            Ok(res) => res
                .error_for_status()
                .or(Err(CheckDomainError::ServiceDNSRecordsNotSet))?
                .text()
                .await
                .or(Err(CheckDomainError::ServiceDNSRecordsNotSet))?,
        };

        if body != token {
            return Err(CheckDomainError::ServiceDNSRecordsNotSet);
        }

        Ok(())
    }
}
