mod proxied_domain;

use crate::domain;
use serde::{Deserialize, Serialize};
use serde_with::{serde_as, TryFromInto};
use std::{collections, net, path, str::FromStr};

fn default_resolver_addr() -> net::SocketAddr {
    net::SocketAddr::from_str("1.1.1.1:53").unwrap()
}

fn default_interface_domain() -> Option<domain::Name> {
    Some(domain::Name::from_str("localhost").unwrap())
}

#[derive(Clone, Deserialize, Serialize)]
pub struct ConfigDNS {
    #[serde(default = "default_resolver_addr")]
    pub resolver_addr: net::SocketAddr,
}

impl Default for ConfigDNS {
    fn default() -> Self {
        Self {
            resolver_addr: default_resolver_addr(),
        }
    }
}

#[derive(Clone, Deserialize, Serialize)]
pub struct ConfigACME {
    pub contact_email: String,
}

#[derive(Clone, Deserialize, Serialize)]
pub struct ConfigBuiltinDomain {
    #[serde(flatten)]
    pub settings: domain::Settings,

    #[serde(default)]
    pub public: bool,
}

#[serde_as]
#[derive(Clone, Deserialize, Serialize)]
pub struct ConfigProxiedDomain {
    #[serde_as(as = "Option<TryFromInto<String>>")]
    pub gemini_url: Option<proxied_domain::GeminiUrl>,

    #[serde_as(as = "Option<TryFromInto<String>>")]
    pub http_url: Option<proxied_domain::HttpUrl>,

    #[serde(default)]
    #[serde_as(as = "TryFromInto<Vec<proxied_domain::HttpRequestHeader>>")]
    pub http_request_headers: proxied_domain::HttpRequestHeaders,

    #[serde(default)]
    pub https_disabled: bool,
}

#[serde_as]
#[derive(Clone, Deserialize, Serialize)]
pub struct ConfigExternalDomain {
    pub tls_key_path: path::PathBuf,
    pub tls_cert_path: path::PathBuf,
}

#[derive(Clone, Deserialize, Serialize)]
pub struct Config {
    pub store_dir_path: path::PathBuf,
    #[serde(default)]
    pub dns: ConfigDNS,
    pub acme: Option<ConfigACME>,

    #[serde(default)]
    pub builtin_domains: collections::HashMap<domain::Name, ConfigBuiltinDomain>,

    #[serde(default)]
    pub proxied_domains: collections::HashMap<domain::Name, ConfigProxiedDomain>,

    #[serde(default = "default_interface_domain")]
    pub interface_domain: Option<domain::Name>,

    #[serde(default)]
    pub external_domains: collections::HashMap<domain::Name, ConfigExternalDomain>,
}
